// EquityLinkedNote.cpp
#include "EquityLinkedNote.h"
#include <algorithm>

namespace QuantLibCppNative {
	// implementations for equity-linked note methods
	EquityLinkedNote::EquityLinkedNote(Real notional, Real initialFixing, std::vector<Date> fixingDates,
		std::vector<Date> paymentDates, Real cap, Real floor)
		: notional_(notional), initialFixing_(initialFixing), fixingDates_(fixingDates),
		paymentDates_(paymentDates), cap_(cap), floor_(floor) {
		// ctor
	}

	bool EquityLinkedNote::isExpired() const {
		Date valuationDate = Settings::instance().evaluationDate();
		// note is expired, if valuation date is greater than the last fixing date
		if (valuationDate > fixingDates_.back())
			return true;
		return false;
	}

	void EquityLinkedNote::setupArguments(PricingEngine::arguments* args) const {
		EquityLinkedNote::arguments* args_ = dynamic_cast<EquityLinkedNote::arguments*>(args);
		QL_REQUIRE(args_ != nullptr, "arguments casting error");
		args_->notional = notional_;
		args_->initialFixing = initialFixing_;
		args_->fixingDates = fixingDates_;
		args_->paymentDates = paymentDates_;
		args_->cap = cap_;
		args_->floor = floor_;
	}

	void EquityLinkedNote::arguments::validate() const {
		// checks for some general argument properties
		QL_REQUIRE(notional > 0.0, "notional must be greater than zero");
		QL_REQUIRE(initialFixing > 0.0, "initial fixing must be greater than zero");
		QL_REQUIRE(cap > floor, "cap must be greater than floor");
		// check for date consistency : all payment dates have to be included 
		// within a set of given fixing dates
		for (int i = 0; i != paymentDates.size(); ++i) {
			if (std::find(fixingDates.begin(), fixingDates.end(), paymentDates[i]) == fixingDates.end()) {
				QL_REQUIRE(false, "payment date has to be included within given fixing dates");
			}
		}
	}

	// implementations for equity-linked path pricer methods
	EquityLinkedNotePathPricer::EquityLinkedNotePathPricer(Real notional, Real initialFixing, std::vector<Date> fixingDates,
		std::vector<Date> paymentDates, Real cap, Real floor, const Handle<YieldTermStructure>& curve)
		: notional_(notional), initialFixing_(initialFixing), fixingDates_(fixingDates),
		paymentDates_(paymentDates), cap_(cap), floor_(floor), curve_(curve) {
		// ctor
	}

	// the actual pricing algorithm for a simulated path is implemented in this method
	Real EquityLinkedNotePathPricer::operator()(const Path& path) const {
		Real coupon = 0.0;
		Real cumulativeCoupon = 0.0;
		Real aggregatePathPayoff = 0.0;
		int paymentDateCounter = 0;

		// loop through fixing dates
		for (int i = 1; i != fixingDates_.size(); ++i) {
			// calculate floating coupon, based on simulated index fixing values
			coupon = std::min(path.at(i) / path.at(i - 1) - 1, cap_);
			// add floating coupon to cumulative coupon
			cumulativeCoupon += coupon;

			// calculate period payoff for each payment date
			if (fixingDates_[i] == paymentDates_[paymentDateCounter]) {
				// calculate discounted payoff for current period, add value to aggregate path payoff
				aggregatePathPayoff += std::max(cumulativeCoupon, floor_) * notional_ * curve_->discount(fixingDates_[i]);
				// re-initialize cumulative coupon to zero, look for the next payment date
				cumulativeCoupon = 0.0;
				paymentDateCounter++;
			}
		}
		return aggregatePathPayoff;
	}
}