// EquityLinkedNote.h
#pragma once
#include <ql/quantlib.hpp>
using namespace QuantLib;

namespace QuantLibCppNative {

	// instrument implementation for equity-linked note
	class EquityLinkedNote : public Instrument {
	public:
		// forward class declarations
		class arguments;
		class engine;

		// ctor and implementations for required base class methods
		EquityLinkedNote(Real notional, Real initialFixing, std::vector<Date> fixingDates,
			std::vector<Date> paymentDates, Real cap, Real floor);
		bool isExpired() const;

	private:
		void setupArguments(PricingEngine::arguments* args) const;
		// term sheet information
		Real notional_;
		Real initialFixing_;
		std::vector<Date> fixingDates_;
		std::vector<Date> paymentDates_;
		Real cap_;
		Real floor_;
	};

	// inner arguments class
	class EquityLinkedNote::arguments : public PricingEngine::arguments{
	public:
		void validate() const;
		Real notional;
		Real initialFixing;
		std::vector<Date> fixingDates;
		std::vector<Date> paymentDates;
		Real cap;
		Real floor;
	};

	// inner engine class
	class EquityLinkedNote::engine
		: public GenericEngine < EquityLinkedNote::arguments, EquityLinkedNote::results > {
		// base class for all further engine implementations
	};

	// path pricer implementation for equity-linked note
	class EquityLinkedNotePathPricer : public PathPricer < Path > {
	public:
		EquityLinkedNotePathPricer(Real notional, Real initialFixing, std::vector<Date> fixingDates,
			std::vector<Date> paymentDates, Real cap, Real floor, const Handle<YieldTermStructure>& curve);
		Real operator()(const Path& path) const;
	private:
		Real notional_;
		Real initialFixing_;
		std::vector<Date> fixingDates_;
		std::vector<Date> paymentDates_;
		Real cap_;
		Real floor_;
		Handle<YieldTermStructure> curve_;
	};

	// monte carlo framework engine implementation for base engine class
	template <typename RNG = PseudoRandom, typename S = Statistics>
	class EquityLinkedNoteMonteCarloPricer : public EquityLinkedNote::engine, private McSimulation < SingleVariate, RNG, S > {
	public:
		// ctor
		EquityLinkedNoteMonteCarloPricer(const boost::shared_ptr<StochasticProcess>& process,
			const Handle<YieldTermStructure>& curve, bool antitheticVariate, Size requiredSamples,
			Real requiredTolerance, Size maxSamples, BigNatural seed)
			: McSimulation<SingleVariate, RNG, S>(antitheticVariate, false), process_(process), curve_(curve),
			requiredSamples_(requiredSamples), requiredTolerance_(requiredTolerance), maxSamples_(maxSamples), seed_(seed) {
			// register observer (pricer) with observables (stochastic process, curve)
			registerWith(process_);
			registerWith(curve_);
		}

		// implementation for required base engine class method
		void calculate() const {
			// the actual simulation process will be performed within the following method
			McSimulation<SingleVariate, RNG, S>::calculate(requiredTolerance_, requiredSamples_, maxSamples_);
			this->results_.value = this->mcModel_->sampleAccumulator().mean();
			//
			if (RNG::allowsErrorEstimate) {
				this->results_.errorEstimate = this->mcModel_->sampleAccumulator().errorEstimate();
			}
			else {
				this->results_.errorEstimate = Null<Real>();
			}
		}

	private:
		// type definitions
		typedef McSimulation<SingleVariate, RNG, S> simulation;
		typedef typename simulation::path_pricer_type path_pricer_type;
		typedef typename simulation::path_generator_type path_generator_type;

		// implementation for McSimulation class virtual method
		TimeGrid timeGrid() const {
			// create time grid based on a set of given index fixing dates
			Date referenceDate = curve_->referenceDate();
			DayCounter dayCounter = curve_->dayCounter();
			std::vector<Time> fixingTimes(arguments_.fixingDates.size());
			for (Size i = 0; i != fixingTimes.size(); ++i) {
				fixingTimes[i] = dayCounter.yearFraction(referenceDate, arguments_.fixingDates[i]);
			}
			return TimeGrid(fixingTimes.begin(), fixingTimes.end());
		}

		// implementation for McSimulation class virtual method
		boost::shared_ptr<path_generator_type> pathGenerator() const {
			// create time grid and get information concerning number of simulation steps for a path
			TimeGrid grid = timeGrid();
			Size steps = (grid.size() - 1);
			// create random sequence generator and return path generator
			typename RNG::rsg_type generator = RNG::make_sequence_generator(steps, seed_);
			return boost::make_shared<path_generator_type>(process_, grid, generator, false);
		}

		// implementation for McSimulation class virtual method
		boost::shared_ptr<path_pricer_type> pathPricer() const {
			// create path pricer implementation for equity-linked note
			return boost::make_shared<EquityLinkedNotePathPricer>(arguments_.notional, arguments_.initialFixing,
				arguments_.fixingDates, arguments_.paymentDates, arguments_.cap, arguments_.floor, this->curve_);
		}

		// private data members
		boost::shared_ptr<StochasticProcess> process_;
		Handle<YieldTermStructure> curve_;
		Size requiredSamples_;
		Real requiredTolerance_;
		Size maxSamples_;
		BigNatural seed_;
	};
}